import logging
import sys

from newspaper import Article as NewspaperArticle

from model.article import Article

logger = logging.getLogger(__name__)

def extract_article(article_url: str) -> Article:
    article = NewspaperArticle(article_url)
    article.download()
    try:
        return _parse_article(article_url, article)
    except LookupError:
        # Newspaper needs some tokenizers from nltk.
        # Download them and try again
        import nltk
        nltk.download('punkt_tab')
        return _parse_article(article_url, article)


def _parse_article(article_url: str, news_article: NewspaperArticle) -> Article:
    news_article.parse()
    news_article.nlp()
    return Article(
        title=news_article.title,
        text=news_article.text,
        keywords=news_article.keywords,
        image_url=news_article.top_image,
        video_url=news_article.movies[0] if len(news_article.movies) > 0 else None,
        url=article_url,
        language=news_article.meta_lang,
    )


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(message)s",
        style="%",
        datefmt="%Y-%m-%d %H:%M",
        level=logging.INFO,
    )
    url = sys.argv[1]
    result = extract_article(url)
    logger.info(result)
    logger.info(result.keywords)



import asyncio

from ai.llm_summarize import summarize_text


async def summarize_batch(
        texts: list[str],
        target_length: int = 500,
        item_length: int = 75,
        article_separator: str = '. '
) -> str:
    summaries = await asyncio.gather(*[summarize_text(text, item_length) for text in texts])
    return article_separator.join(summaries)



from ai.llm_translate import translate_text
from model.article import Article
from model.article_version import ArticleVersion, TranslatedArticle
from transform.article_transformer import ArticleTransformer


class ArticleTranslator(ArticleTransformer):

    def __init__(self, target_language: str):
        self.target_language = target_language

    async def transformed_article(self, article: Article) -> ArticleVersion:
        """
        Returns the same contents in a different language (target_language)
        """
        translated_article: Article = article.copy()
        translated_article.text = await translate_text(article.text, self.target_language)
        return TranslatedArticle(
            language=self.target_language,
            article=translated_article,
            parent_article=article,
        )



import os


def create_config_dir() -> bool:
    """
    Creates a directory for the application in the user home
    (os-independent)
    :return True, if a directory was created, otherwise False
    """
    config_dir = os.path.expanduser('~/.plainews/')
    if not os.path.exists(config_dir):
        os.mkdir(config_dir)
        return True
    else:
        return False



from ai.llm_summarize import summarize_text
from model.article import Article
from model.article_version import ArticleVersion, SummarizedArticle
from transform.article_transformer import ArticleTransformer


class ArticleSummary(ArticleTransformer):

    def __init__(self, target_length: int):
        self.target_length = target_length

    async def transformed_article(self, article: Article) -> ArticleVersion:
        """
        Returns a more concise version of the source article
        """
        short_article = article.copy()
        short_article.text = await summarize_text(article.text, self.target_length)
        return SummarizedArticle(
            parent_article=article,
            article=short_article,
            target_word_length=self.target_length,
        )



from textual import on
from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import OptionList, Button

from ui.latest_articles_screen import LatestArticlesScreen


class FeedsList(Widget):

    CSS_PATH = ['feeds_list.tcss']
    OPTIONS_LIST_ID = 'feeds_option_list'

    def __init__(
            self,
            *children: Widget,
            name: str | None = None,
            id: str | None = None,
            classes: str | None = None,
            disabled: bool = False,
            feed_items: list[str] | None = None,
    ):
        super().__init__(*children, name=name, id=id, classes=classes, disabled=disabled)
        self._feed_items = feed_items

    def compose(self) -> ComposeResult:
        yield OptionList(*self._feed_items, id=self.OPTIONS_LIST_ID)
        yield Button('Since you were last here', id='latest_button')

    @on(Button.Pressed, selector='#latest_button')
    def on_latest_articles_pressed(self, button: Button):
        articles_screen = LatestArticlesScreen()
        self.app.push_screen(articles_screen)

    @property
    def feed_items(self) -> list[str]:
        return self._feed_items

    @feed_items.setter
    def feed_items(self, new_value: list[str]):
        self._feed_items = new_value
        self._options_list().clear_options()
        self._options_list().add_options(new_value)

    def _options_list(self) -> OptionList:
        return self.query_exactly_one(selector=f'#{FeedsList.OPTIONS_LIST_ID}', expect_type=OptionList)



import os

from litellm import acompletion

from config.config import Config

config = Config()

# There doesn't seem to be a way to specify this as part of invoking the
# completion function - it needs to be set as an ENV var
os.environ["AZURE_API_TYPE"] = config.llm_api_type if config.llm_api_type else ""

async def llm_completion(system_prompt: str, payload: str) -> str:
    response = await acompletion(
        model=_get_llm_model(),
        messages=_get_messages(system_prompt, payload),
        api_base=config.llm_base_url,
        api_key=config.llm_api_key,
        api_version=config.llm_version,
    )
    return response.choices[0].message.content


def _get_llm_model() -> str:
    return config.llm_model


def _get_messages(system_prompt: str, payload: str) -> list[dict]:
    # Seems that Ollama gets confused if chat history with different roles is provided. This is
    # probably not the case with more powerful models
    return [
        {"content": system_prompt + '\n\n' + payload, "role": "assistant"},
    ]
    # return [
    #     {"content": system_prompt, "role": "assistant"},
    #     {"content": payload, "role": "user"}
    # ]



from textual.app import App

from ui.home import HomeScreen


class Plainews(App[None]):

    async def on_mount(self) -> None:
        await self.push_screen(HomeScreen())


def run_app():
    app = Plainews()
    app.run()


if __name__ == '__main__':
    run_app()



import os
import signal
import subprocess

from config.config import Config


class TextToSpeech:
    """
    Base class for all text-to-speech implementations
    """

    def speak_text(self, text: str):
        pass

    def cancel(self):
        pass

    def is_playing(self) -> bool:
        pass

    def is_paused(self) -> bool:
        pass

    def pause(self):
        pass

    def resume(self):
        pass


class CmdTextToSpeech(TextToSpeech):
    """
    Implements TTS by running a local script, provided by a config option. It allows
    users to specify a working TTS command that they have installed locally that will
    generate and play audio for the text. For instance, in MacOS, the `say` cmd utility
    can be used for a single and performant TTS
    """

    config = Config()
    text_marker = '<text>'

    def __init__(self):
        super().__init__()
        self.subprocess: subprocess.Popen | None = None
        self.is_process_suspended: bool = False

    def speak_text(self, text: str):
        if not self.config.tts or not self.config.tts.cmd:
            return
        command = self.config.tts.cmd.replace(self.text_marker, text)
        process = subprocess.Popen(command, shell=True)
        self.subprocess = process
        self.is_process_suspended = False

    def cancel(self):
        if self.subprocess and self.subprocess.poll() is None:  # Check if the process is still running
            self.subprocess.terminate()  # Terminate the process
            self.subprocess.wait()  # Wait for the process to terminate
            self.is_process_suspended = False

    def is_playing(self) -> bool:
        return self.subprocess and self.subprocess.poll() is None

    def pause(self):
        if self.subprocess and self.subprocess.poll() is None:
            os.kill(self.subprocess.pid, signal.SIGSTOP)
            self.is_process_suspended = True

    def resume(self):
        if self.subprocess and self.subprocess.poll() is None:
            os.kill(self.subprocess.pid, signal.SIGCONT)
            self.is_process_suspended = False

    def is_paused(self) -> bool:
        """
        Checks if there's a suspended speech command
        :return: Yes, if playback can be resumed
        """
        # also checks for an existing subprocess that hasn't exited yet as a double check
        return self.subprocess and self.subprocess.poll() is None and self.is_process_suspended


def has_tts_setup() -> bool:
    """
    Determines if the application has text-to-speech configured (based on the configuration).
    :return: True, if the config contains options for using TTS, or False otherwise
    """
    config = Config()
    return config.tts is not None


def speak_text(text: str) -> TextToSpeech:
    """
    Uses the configured TTS options in order to say the provided text
    :param text: The text to say
    :return: A TextToSpeech instance that can be used to stop the audio
    """
    # so far, only TTS based on a custom command is supported
    tts = CmdTextToSpeech()
    tts.speak_text(text)
    return tts



import webbrowser

from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import TextArea, Static
from textual import work

from ai.tts import speak_text, has_tts_setup, TextToSpeech
from config.config import Config
from model.article import Article
from model.feed import FeedEntry
from model.feed_manager import FeedManager
from transform.article_summary import ArticleSummary
from transform.article_transformer import ArticleTransformer
from transform.article_translator import ArticleTranslator


class ArticleView(Widget):
    BINDINGS = [
        ("o", "open_in_browser", "Open in browser"),
        ("t", "translate_article", "Translate"),
        ("s", "summarize_article", "Summarize"),
        ("a", "speak_article", "Audio"),
        ("p", "pause", "Pause"),
        ("c", "resume", "Resume"),
    ]

    title = reactive("No article selected")
    text = reactive('Select an item on the left')

    config = Config()
    feed_manager = FeedManager()

    def __init__(
            self,
            entry: FeedEntry | None = None,
            selected_article: Article | None = None,
            id: str | None = None
    ):
        super().__init__(id=id)
        self.tts: TextToSpeech | None = None
        self.selected_entry = entry
        base_article = entry.meta.base_article if entry else None
        self.selected_article = selected_article if selected_article else base_article
        self.title = self.selected_article.title if self.selected_article else ""
        self.text = self.selected_article.text if self.selected_article else ""

    def update_text(self):
        title = self.query_one(Static)
        title.update(self.title)
        body = self.query_one(TextArea)
        body.text = self.text

    def compose(self) -> ComposeResult:
        yield Static(self.title, id='article-title')
        area = TextArea(self.text, id='article-text')
        area.read_only = True
        yield area

    def action_open_in_browser(self):
        if self.selected_entry:
            webbrowser.open(self.selected_entry.link)

    async def action_translate_article(self):
        self.notify("Translating article. Please wait...")
        self._translate_article()

    async def action_summarize_article(self):
        self.notify("Summarizing. Please wait...")
        self._summarize_article()

    async def action_speak_article(self):
        if self.tts and self.tts.is_playing():
            self.tts.cancel()
        else:
            if self.selected_article:
                self.tts = speak_text(self.selected_article.text)
        self.refresh_bindings()

    async def action_pause(self):
        if self.tts:
            self.tts.pause()
            self.refresh_bindings()

    async def action_resume(self):
        if self.tts:
            self.tts.resume()
            self.refresh_bindings()

    def check_action(
            self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:
        actions_while_tts = ['pause']
        actions_while_reading = list(
            map(lambda m: m[1],
                filter(lambda b: b[1] not in actions_while_tts, self.BINDINGS))
        )
        if action == 'open_in_browser':
            return self.selected_entry is not None and self.selected_entry.link is not None
        if action == 'resume':
            return self.tts is not None and self.tts.is_paused()
        if action == 'speak_article':
            return has_tts_setup()
        if self.tts and self.tts.is_playing():
            return action in actions_while_reading + actions_while_tts
        else:
            return action in actions_while_reading

    @work(exclusive=True, exit_on_error=False)
    async def _translate_article(self):
        translator = ArticleTranslator(self.config.language)
        await self._apply_transformer(translator)

    @work(exclusive=True, exit_on_error=False)
    async def _summarize_article(self):
        summarizer = ArticleSummary(self.config.summarization_target_length)
        await self._apply_transformer(summarizer)

    def _update_view(self, article: Article):
        self.title = article.title
        self.text = article.text
        self.update_text()

    async def _apply_transformer(self, transformer: ArticleTransformer):
        if self.selected_entry:
            version = await self.feed_manager.create_article_version(self.selected_entry, transformer)
            self.selected_article = version.article
        else:
            new_version = await transformer.transformed_article(self.selected_article)
            self.selected_article = new_version.article
        self._update_view(self.selected_article)



from pydantic import BaseModel


class Article(BaseModel):
    title: str
    text: str
    url: str
    keywords: list[str]
    image_url: str | None = None
    video_url: str | None = None
    language: str

    def __str__(self) -> str:
        return f"{self.title}: ({self.text[:30]})"



from datetime import datetime

from pydantic import BaseModel

class Session(BaseModel):
    """
    Contains app sessions - not to be confused with the app config
    that's stored as a yaml configuration file. A session contains
    state variables that the application generates and stores.
    """
    last_opened: datetime | None = None

    """
    Last time feeds were refreshed (successfully)
    """
    last_feed_refresh: datetime | None = None



import functools
from datetime import datetime

from model.article import Article
from model.feed import FeedEntry
from model.feed_manager import FeedManager
from model.feed_pipeline import FeedPipeline
from transform.article_summary import ArticleSummary


def articles_since(since_date: datetime, force_fetch: bool) -> list[FeedEntry]:
    feed_manager = FeedManager()
    if force_fetch or feed_manager.outdated_feeds():
        feed_manager.refresh_feeds()
        feed_manager.session_manager.update_last_feed_refresh(datetime.now())
    feeds = feed_manager.get_feeds()
    entries = functools.reduce(lambda a, b: a + b.entries, feeds, [])
    return list(filter(lambda e: since_date < e.publish_date, entries))


async def latest_articles_summary(since_date: datetime, force_fetch: bool) -> Article:
    entries = articles_since(since_date, force_fetch)
    return await articles_summary(entries)


async def articles_summary(entries: list[FeedEntry], delimiter: str = '\n\n-------------\n\n') -> Article:
    pipeline = _latest_articles_pipeline(entries)
    articles = await pipeline.transform()
    article_texts = [article.text for article in articles]
    summary = delimiter.join(article_texts)
    return Article(
        title=f"Summary of the selected articles",
        text=summary,
        language='english',
        url='',
        keywords=[],
        video_url=None,
        image_url=None,
    )


def _latest_articles_pipeline(entries: list[FeedEntry]) -> FeedPipeline:
    return FeedPipeline(entries, transformers=[
        ArticleSummary(150)
    ])



import sys
from datetime import datetime
from time import mktime

import feedparser

from model.feed import Feed, FeedEntry


def parse_rss_feed(feed_url: str) -> Feed:
    rss = feedparser.parse(feed_url)
    title = rss.feed.title
    subtitle = rss.feed.subtitle if 'subtitle' in rss.feed else None
    entries = []
    for entry in rss.entries:
        entry_title = entry.title
        entry_link = entry.link
        entry_id = entry.id
        publish_date = datetime.fromtimestamp(mktime(entry.published_parsed))
        update_date = datetime.fromtimestamp(mktime(entry.updated_parsed))
        summary = entry.summary
        feed_entry = FeedEntry(
            entry_id=entry_id,
            title=entry_title,
            link=entry_link,
            summary=summary,
            publish_date=publish_date,
            update_date=update_date
        )
        entries.append(feed_entry)
    return Feed(
        title=title,
        subtitle=subtitle,
        entries=sorted(entries, key=lambda e: e.update_date, reverse=True),
        url=feed_url,
    )


if __name__ == "__main__":
    rss_url = sys.argv[1]
    feed = parse_rss_feed(rss_url)
    print(f"Parsed feed: {feed}")



import os

from pydantic import BaseModel
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource, YamlConfigSettingsSource

APP_CONFIG_DIR = '~/.plainews'

class HistoryConfig(BaseModel):
    keep_history: bool | None = False
    max_history_items_per_feed: int | None = 150


class TextToSpeechConfig(BaseModel):
    cmd: str | None = None


class Config(BaseSettings):
    llm_model: str = "ollama/llama3.1"
    llm_base_url: str | None = None   # "http://localhost:11434" for Ollama
    llm_api_key: str | None = None
    llm_api_type: str | None = None   # needed for Azure OpenAI
    llm_version: str | None = None    # needed for Azure OpenAI
    language: str = "english"
    feeds_file: str = os.path.expanduser(f'{APP_CONFIG_DIR}/feeds.pickle')
    session_file: str = os.path.expanduser(f'{APP_CONFIG_DIR}/session.pickle')
    transformers: list[str] = []
    followed_feeds: list[str] = [
        "https://rss.sueddeutsche.de/rss/Topthemen",
        "https://feeds.arstechnica.com/arstechnica/index",
    ]
    summarization_target_length: int | None = 250
    history: HistoryConfig = HistoryConfig()
    tts: TextToSpeechConfig | None = None

    # There are two options - reading the config from a file in the PWD, or reading
    # a file from the home directory
    model_config = SettingsConfigDict(
        yaml_file=[
            os.path.expanduser('~/.plainews.yml'),
            os.path.expanduser(f'{APP_CONFIG_DIR}/config.yml'),
            'config.yml'
        ]
    )

    @classmethod
    def settings_customise_sources(
            cls,
            settings_cls: type[BaseSettings],
            init_settings: PydanticBaseSettingsSource,
            env_settings: PydanticBaseSettingsSource,
            dotenv_settings: PydanticBaseSettingsSource,
            file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """
        This overridden method is needed in order to include the Yaml source
        into the list while initializing
        :param settings_cls:
        :param init_settings:
        :param env_settings:
        :param dotenv_settings:
        :param file_secret_settings:
        :return:
        """
        return (
            YamlConfigSettingsSource(settings_cls),
            init_settings,
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )



import os
import pickle
from config.config import Config
from model.feed import Feed
from storage.config_dir import create_config_dir


class StoredFeeds:

    config = Config()

    def __init__(self, feeds: list[Feed]):
        self.feeds = feeds

    def add_feed(self, feed: Feed):
        self.feeds.append(feed)
        self.save_feeds()

    def set_feeds(self, feeds: list[Feed]):
        self.feeds = feeds
        self.save_feeds()

    def remove_feed(self, feed: Feed):
        self.feeds.remove(feed)
        self.save_feeds()

    def remove_feeds(self):
        self.feeds = []
        self.save_feeds()

    def get_feed(self, url: str) -> Feed | None:
        for feed in self.feeds:
            if feed.url == url:
                return feed
        return None

    def merge_feed(self, feed: Feed) -> Feed:
        """
        Adds the feed to the storage. If the feed already exists, merges the
        already stored entries with any new ones from the `feed` parameter. Stores the
        merged entries in chronological order
        :param feed: the feed to merge into the storage
        :return: the (possibly modified) feed that was stored
        """
        existing_feed = self.get_feed(feed.url)
        if not existing_feed:
            self.add_feed(feed)
            return feed
        else:
            merged = self._merged_feed(existing_feed, feed)
            self.replace_feed(existing_feed, merged)
            return merged

    def replace_feed(self, old_feed: Feed, new_feed: Feed):
        """
        As opposed to removing and adding a feed, this method keeps the
        order in the feed list
        """
        self.feeds = [feed if feed.url != old_feed.url else new_feed for feed in self.feeds]
        self.save_feeds()

    def save_feeds(self):
        try:
            with open(self.config.feeds_file, 'wb') as storage_file:
                pickle.dump(self.feeds, storage_file)
        except FileNotFoundError:
            if create_config_dir():
                # retry (caution - it's recursing)
                self.save_feeds()

    def load_feeds(self):
        if os.path.exists(self.config.feeds_file):
            with open(self.config.feeds_file, 'rb') as feeds_file:
                self.feeds = pickle.load(feeds_file)

    @staticmethod
    def _merged_feed(existing_feed: Feed, new_feed: Feed) -> Feed:
        merged_set = set(existing_feed.entries)
        merged_set.update(new_feed.entries)
        return Feed(
            entries=sorted(list(merged_set), key=lambda e: e.update_date, reverse=True),
            url=new_feed.url,
            title=new_feed.title,
            subtitle=new_feed.subtitle,
        )



from config.config import Config
from model.article import Article
from transform.article_summary import ArticleSummary
from transform.article_transformer import ArticleTransformer
from transform.article_translator import ArticleTranslator


class TransformationBuilder:

    config = Config()
    transformer_mapping = {
        "ArticleTranslator": ArticleTranslator(config.language),
        "ArticleSummary": ArticleSummary()
    }

    def __init__(self, transformers: list[ArticleTransformer]):
        self.transformers = transformers

    async def transform(self, article: Article) -> Article:
        result = article
        for transformer in self.transformers:
            result = transformer.transformed_article(result)
        return result

    @staticmethod
    def from_names(transformer_names: list[str]) -> "TransformationBuilder":
        return TransformationBuilder(
            list(map(lambda name: TransformationBuilder.transformer_mapping[name], transformer_names))
        )

    @staticmethod
    def from_config() -> "TransformationBuilder":
        return TransformationBuilder(
            list(map(
                lambda name: TransformationBuilder.transformer_mapping[name],
                TransformationBuilder.config.transformers)
            )
        )



from model.article import Article
from model.feed import FeedEntry
from model.feed_manager import FeedManager
from transform.article_transformer import ArticleTransformer


class FeedPipeline:

    feed_manager = FeedManager()

    def __init__(self, entries: list[FeedEntry], transformers: list[ArticleTransformer]):
        self.entries = entries
        self.transformers = transformers

    async def transform(self) -> list[Article]:
        articles = [self.feed_manager.extract_article(entry) for entry in self.entries]
        transformed = [await self._apply_transformations(article) for article in articles]
        return transformed

    async def _apply_transformations(self, article: Article) -> Article:
        transformed_article = article
        for trans in self.transformers:
            new_version = await trans.transformed_article(transformed_article)
            transformed_article = new_version.article
        return transformed_article



from datetime import datetime

from pydantic import BaseModel

from model.article_version import ArticleVersion


class FeedEntryMeta(BaseModel):
    """
    Contains metadata attached not from the RSS feed, but from the app.
    For instance, if the entry has been read, a possible generated Article
    object from it etc.
    """
    is_read: bool = False
    base_article: ArticleVersion | None = None
    generated_articles: list[ArticleVersion] = []

    def add_article_version(self, version: ArticleVersion):
        to_replace: ArticleVersion | None = None
        for generated_version in self.generated_articles:
            if isinstance(generated_version, version.__class__):
                to_replace = generated_version
        if to_replace:
            self.generated_articles.remove(to_replace)
        self.generated_articles.append(version)


class FeedEntry(BaseModel):
    entry_id: str
    title: str
    link: str
    publish_date: datetime
    update_date: datetime
    summary: str

    meta: FeedEntryMeta = FeedEntryMeta()

    def __hash__(self):
        return hash(self.entry_id)

    def __eq__(self, other):
        if not isinstance(other, FeedEntry):
            return NotImplemented
        return self.entry_id == other.entry_id

class Feed(BaseModel):
    url: str
    title: str
    subtitle: str | None
    entries: list[FeedEntry]



from textual import on
from textual.app import ComposeResult
from textual.screen import ModalScreen

from ui.latest_articles_view import LatestArticlesView


class LatestArticlesScreen(ModalScreen):

    CSS_PATH = ['latest_articles_view.tcss']

    def compose(self) -> ComposeResult:
        yield LatestArticlesView()

    @on(LatestArticlesView.Escape)
    def on_pop_screen(self):
        self.app.pop_screen()



from model.article import Article
from model.article_version import ArticleVersion


class ArticleTransformer:

    async def transformed_article(self, article: Article) -> ArticleVersion:
        """
        Performs a transformation of the article - it post-processes the input
        and brings a more refined version (in a different language, just a summary etc.)
        :param article: the raw article
        :return: a new ArticleVersion object that has the target transformation applied to it
        """
        pass



from ai.llm import llm_completion

TARGET_LENGTH_TOKEN = "{target_length}"
SUMMARIZE_PROMPT = f"""Summarize the following article. Mention all the important information inside and ignore filler 
information and unrelated information. The summary should be roughly around {TARGET_LENGTH_TOKEN} words"""

async def summarize_text(text: str, target_length: int) -> str:
    prompt = SUMMARIZE_PROMPT.replace(TARGET_LENGTH_TOKEN, str(target_length))
    return await llm_completion(prompt, text)



from model.feed_manager import FeedManager


def test_storing_feed():
    feed_manager = FeedManager()
    feed_manager.add_feed("https://rss.sueddeutsche.de/rss/Topthemen")
    feeds = feed_manager.get_feeds()
    assert len(feeds) == 1



import os
import pickle

from config.config import Config
from model.session import Session
from storage.config_dir import create_config_dir


class StoredSession:

    config = Config()

    def __init__(self):
        self.session: Session | None = None
        self.load_session()

    def save_session(self, new_session: Session):
        self.session = new_session
        try:
            with open(self.config.session_file, 'wb') as session_file:
                pickle.dump(self.session, session_file)
        except FileNotFoundError:
            if create_config_dir():
                # retry (caution - it's recursing)
                self.save_session(new_session)

    def load_session(self) -> Session | None:
        if os.path.exists(self.config.session_file):
            with open(self.config.session_file, 'rb') as session_file:
                self.session = pickle.load(session_file)
                return self.session
        return None



from textual.app import ComposeResult
from textual.events import Key
from textual.screen import Screen
from textual.widgets import Footer

from model.article import Article
from model.feed import FeedEntry
from ui.article_view import ArticleView


class ArticleScreen(Screen):

    def __init__(
            self,
            feed_entry: FeedEntry | None = None,
            article: Article | None = None,
            name: str | None = None,
            id: str | None = None,
            classes: str | None = None,
    ):
        super().__init__(name, id, classes)
        self.article = article
        self.feed_entry = feed_entry

    def compose(self) -> ComposeResult:
        article_view = ArticleView(self.feed_entry, self.article)
        yield article_view
        yield Footer()

    def on_key(self, event: Key) -> None:
        if event.key == 'q' or event.key == 'escape':
            self.app.pop_screen()



from ai.llm import llm_completion

LANGUAGE_TOKEN = "{language}"
TRANSLATE_PROMPT = f"""
Translate the following text to {LANGUAGE_TOKEN}. Don't paraphrase or change the text in any other way. Keep the 
formatting."""

async def translate_text(text: str, language: str) -> str:
    prompt = TRANSLATE_PROMPT.replace(LANGUAGE_TOKEN, language)
    return await llm_completion(prompt, text)



import logging
import webbrowser
from datetime import datetime
from typing import cast

from textual import on
from textual import work
from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import Header, Footer, OptionList
from textual.widgets.option_list import Option
from textual.widgets.option_list import Separator
from textual.worker import Worker, WorkerState

from ai.tts import speak_text, TextToSpeech, has_tts_setup
from config.config import Config
from model.article import Article
from model.feed import Feed, FeedEntry
from model.feed_manager import FeedManager
from model.session_manager import SessionManager
from transform.article_summary import ArticleSummary
from transform.article_translator import ArticleTranslator
from rich.text import Text
from ui.article_view import ArticleView
from ui.articles_list import ArticlesList
from ui.feeds_list import FeedsList

logger = logging.getLogger(__name__)


class ReaderView(Widget):

    CSS_PATH = ['reader_view.tcss', 'article_layout.tcss']

    BINDINGS = [
        ("r", "refresh_feeds", "Refresh"),
    ]

    config = Config()
    feed_manager = FeedManager()
    session_manager = SessionManager()

    def __init__(
            self,
            *children: Widget,
            name: str | None = None,
            id: str | None = None,
            classes: str | None = None,
            disabled: bool = False,
    ) -> None:
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )
        self.plainews = cast("Plainews", self.app)
        self.selected_feed: Feed | None = None
        self.selected_entry: FeedEntry | None = None
        self.selected_article: Article | None = None
        self.tts: TextToSpeech | None = None
        session = self.session_manager.session
        self.last_feed_update = session.last_opened

    async def on_mount(self) -> None:
        session = self.session_manager.session
        session.last_opened = datetime.now()
        self.session_manager.save_session(session)
        feeds = self.feed_manager.get_feeds()
        self._update_feeds_list(feeds)
        self._refresh_feeds()

    def compose(self) -> ComposeResult:
        yield Header()
        yield FeedsList(feed_items=[], id='sidebar-feeds', classes='sidebar-expanded')
        yield ArticlesList(*[], id='sidebar-articles', classes='sidebar-collapsed')
        yield ArticleView()
        yield Footer()

    @on(OptionList.OptionSelected)
    async def feed_selected(self, option: OptionList.OptionSelected):
        if option.option_list is self.feeds_option_list:
            selected_index = option.option_index
            self.selected_feed = self.feed_manager.get_feeds()[selected_index]
            await self.refresh_selected_feed()
        if isinstance(option.option_list, ArticlesList):
            self.selected_entry = self.selected_feed.entries[option.option_index]
            self.selected_article = self.feed_manager.extract_article(self.selected_entry)
            self.selected_entry.meta.is_read = True
            self.update_article_screen(self.selected_article)
            self.feed_manager.save_feeds()
            await self.refresh_selected_feed()

    def action_open_in_browser(self):
        if self.selected_entry:
            webbrowser.open(self.selected_entry.link)

    async def action_translate_article(self):
        if self.selected_entry:
            self.notify("Translating article. Please wait...")
            self._translate_article()

    async def action_summarize_article(self):
        if self.selected_entry:
            self.notify("Summarizing. Please wait...")
            self._summarize_article()

    async def action_refresh_feeds(self):
        self.selected_feed = self.feed_manager.refresh_feed(self.selected_feed)
        await self.refresh_selected_feed()

    async def action_speak_article(self):
        if self.tts and self.tts.is_playing():
            self.tts.cancel()
        else:
            if self.selected_article:
                self.tts = speak_text(self.selected_article.text)
        self.refresh_bindings()

    async def action_pause(self):
        if self.tts:
            self.tts.pause()
            self.refresh_bindings()

    async def action_resume(self):
        if self.tts:
            self.tts.resume()
            self.refresh_bindings()

    def check_action(
            self, action: str, parameters: tuple[object, ...]
    ) -> bool | None:
        actions_while_tts = ['pause']
        actions_while_reading = list(
            map(lambda m: m[1],
                filter(lambda b: b[1] not in actions_while_tts, self.BINDINGS))
        )
        if action == 'resume':
            return self.tts is not None and self.tts.is_paused()
        if action == 'speak_article':
            return has_tts_setup()
        if self.tts and self.tts.is_playing():
            return action in actions_while_reading + actions_while_tts
        else:
            return action in actions_while_reading

    @work(exclusive=True, exit_on_error=False, name='refresh_feeds', thread=True)
    async def _refresh_feeds(self) -> list[Feed]:
        return self.feed_manager.refresh_feeds()

    @work(exclusive=True, exit_on_error=False)
    async def _translate_article(self):
        translator = ArticleTranslator(self.config.language)
        version = await self.feed_manager.create_article_version(self.selected_entry, translator)
        self.selected_article = version.article
        self.update_article_screen(version.article)

    @work(exclusive=True, exit_on_error=False)
    async def _summarize_article(self):
        summarizer = ArticleSummary(self.config.summarization_target_length)
        version = await self.feed_manager.create_article_version(self.selected_entry, summarizer)
        self.selected_article = version.article
        self.update_article_screen(version.article)

    async def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        """Called when the worker state changes."""
        if event.state == WorkerState.ERROR:
            self.notify(
                title=self._error_title_for_worker(event.worker),
                message=f"Error: {str(event.worker.error)[:300]}",
                severity="error")
            self.log(event)
        if event.worker.name == 'refresh_feeds' and event.state == WorkerState.SUCCESS:
            feeds = event.worker.result
            self._update_feeds_list(feeds)

    def _article_screen(self) -> ArticleView | None:
        return self.query_one(ArticleView)

    def _update_feeds_list(self, feeds: list[Feed]):
        feeds_titles = [feed.title for feed in feeds]
        self.feeds_option_list.clear_options()
        self.feeds_option_list.add_options(feeds_titles)

    def update_article_screen(self, article: Article):
        screen = self._article_screen()
        screen.selected_entry = self.selected_entry
        screen.selected_article = self.selected_article
        screen.title = article.title
        screen.text = article.text
        screen.update_text()

    async def refresh_selected_feed(self):
        articles_list = self.query_one(ArticlesList)
        articles_list.clear_options()
        articles_list.add_options(self._options_for_feed(self.selected_feed))
        self._restore_selection()

    @property
    def feeds_option_list(self) -> OptionList:
        return self.query_exactly_one(selector=f'#{FeedsList.OPTIONS_LIST_ID}', expect_type=OptionList)

    @staticmethod
    def _options_for_feed(feed: Feed) -> list[Option | Separator]:
        """
        Returns the options to be added to the FeedsList, sorted chronologically
        :param feed: the selected Feed whose articles will be listed
        :return: a list of options, split by separators, ready to be displayed by an OptionsList
        """
        options = []
        for entry in feed.entries:
            style = 'italic' if hasattr(entry, 'meta') and entry.meta.is_read else 'bold'
            options.append(Text(entry.title, style=style))
            options.append(Separator())
        return options

    def _restore_selection(self):
        try:
            index = self.selected_feed.entries.index(self.selected_entry)
            article_list: ArticlesList = cast(ArticlesList, self.query_one(ArticleView))
            article_list.highlighted = index
        except ValueError:
            pass

    @staticmethod
    def _error_title_for_worker(worker: Worker) -> str | None:
        titles_dict = {
            '_translate_article': "Translation failed",
            '_summarize_article': "Summarization failed",
        }
        return titles_dict[worker.name] if worker.name in titles_dict else None




from datetime import datetime

from model.session import Session
from storage.stored_session import StoredSession


class SessionManager:

    stored_session = StoredSession()

    def __init__(self):
        self.session = self.stored_session.load_session()
        self.session = self.session if self.session else self._empty_session()

    def save_session(self, session: Session | None = None):
        """
        Persistently stores sessions
        :param session: session DTO - if None, the pre-cached object is saved
        :return:
        """
        self.stored_session.save_session(session)

    def update_last_feed_refresh(self, last_refresh: datetime):
        """
        Updates the timestamp of the last feed refresh
        :param last_refresh: datetime when the feeds were last updated
        """
        self.session = self.stored_session.load_session()
        self.session.last_feed_refresh = last_refresh
        self.save_session(self.session)

    @staticmethod
    def _empty_session() -> Session:
        return Session()



import logging

from textual.app import ComposeResult
from textual.screen import Screen

from ui.reader_view import ReaderView

logger = logging.getLogger(__name__)


class HomeScreen(Screen[None]):

    CSS_PATH = ['home_layout.tcss', 'reader_view.tcss', 'article_layout.tcss', 'feeds_list.tcss',
                'latest_articles_view.tcss']

    def compose(self) -> ComposeResult:
        yield ReaderView()



from datetime import datetime, timedelta

from config.config import Config
from model.article import Article
from model.article_version import ArticleVersion
from model.feed import Feed, FeedEntry
from model.session_manager import SessionManager
from storage.stored_feeds import StoredFeeds
from model.feed_parser import parse_rss_feed
from model.article_extraction import extract_article
from transform.article_transformer import ArticleTransformer


def _article_from_feed_entry(feed_entry: FeedEntry) -> Article:
    article = extract_article(feed_entry.link)
    return article


class FeedManager:

    REFRESH_INTERVAL_SECS = 120

    storage = StoredFeeds([])
    config = Config()
    session_manager = SessionManager()

    def __init__(self):
        # load the already saved feeds from storage. New ones can be fetched using refresh_feeds
        self.storage.load_feeds()

    def add_feed(self, feed_url: str) -> Feed:
        feed = parse_rss_feed(feed_url)
        self.storage.add_feed(feed)
        return feed

    def refresh_feed(self, feed: Feed) -> Feed:
        new_feed = parse_rss_feed(feed.url)
        self._replace_feed(feed, new_feed)
        return new_feed

    def refresh_feeds(self) -> list[Feed]:
        feeds = list(map(lambda feed_url: parse_rss_feed(feed_url), self.config.followed_feeds))
        if self.config.history.keep_history:
            self.storage.load_feeds()
            [self.storage.merge_feed(feed) for feed in feeds]
            return self.get_feeds()
        else:
            self.storage.set_feeds(feeds)
        self.session_manager.update_last_feed_refresh(datetime.now())
        return feeds

    def get_feeds(self) -> list[Feed]:
        return self.storage.feeds

    def remove_feed(self, feed: Feed):
        self.storage.remove_feed(feed)
        self.storage.save_feeds()

    def save_feeds(self):
        self.storage.save_feeds()

    @staticmethod
    def extract_article(feed_entry: FeedEntry) -> Article:
        article = extract_article(feed_entry.link)
        extracted_version = ArticleVersion(
            parent_article=None,
            article=article,
        )
        feed_entry.meta.base_article = article
        feed_entry.meta.add_article_version(extracted_version)
        return article

    async def create_article_version(self, feed_entry: FeedEntry, transformer: ArticleTransformer) -> ArticleVersion:
        base_article = feed_entry.meta.base_article
        if not base_article:
            base_article = self.extract_article(feed_entry)
        transformed = await transformer.transformed_article(base_article)
        feed_entry.meta.add_article_version(transformed)
        self.save_feeds()
        return transformed

    def outdated_feeds(self) -> bool:
        """
        Checks if the feeds need to be refreshed
        :return: True, if feeds are older than REFRESH_INTERVAL_SECS. False, otherwise
        """
        last_refresh = self.session_manager.session.last_feed_refresh if (
            self.session_manager.session.last_feed_refresh) \
            else datetime.min
        if not self.session_manager.session.last_feed_refresh:
            return True
        return last_refresh + timedelta(seconds=self.REFRESH_INTERVAL_SECS) < datetime.now()

    def _replace_feed(self, old_feed: Feed, new_feed: Feed):
        self.storage.replace_feed(old_feed, new_feed)



from datetime import datetime, timedelta
from typing import cast

from textual import events, work
from textual import on
from textual.app import ComposeResult
from textual.containers import Grid
from textual.events import Key
from textual.message import Message
from textual.widget import Widget
from textual.widgets import SelectionList, Label, Button
from textual.worker import Worker, WorkerState

from model.article import Article
from model.feed import FeedEntry
from model.feed_manager import FeedManager
from model.latest_articles import articles_since, articles_summary
from model.session_manager import SessionManager
from ui.article_screen import ArticleScreen


class LatestArticlesView(Widget):

    class Escape(Message):
        pass

    session_manager = SessionManager()
    feed_manager = FeedManager()

    def __init__(self,
                 *children: Widget,
                 name: str | None = None,
                 id: str | None = None,
                 classes: str | None = None,
                 disabled: bool = False,
                 ):
        super().__init__(*children, name=name, id=id, classes=classes, disabled=disabled)
        self.current_items: list[FeedEntry] = []

    def compose(self) -> ComposeResult:
        yield Grid(
            Label('What happened since you were last here:', id='latest_articles_title'),
            SelectionList[int](*[], id='latest_articles_list'),
            Button("Create a summary", variant="primary", id="summary_button"),
            Button("Open selected article", variant="default", id="open_button"),
            id='latest_articles_grid',
        )

    def on_mount(self, event: events.Mount) -> None:
        self.current_items = articles_since(
            self._last_opened_time() - timedelta(days=1), force_fetch=False
        )
        titles = [entry.title for entry in self.current_items]
        self.latest_articles_list.clear_options()
        self.latest_articles_list.add_options(self._selection_list_items(titles))

    def on_key(self, event: Key) -> None:
        if event.key == 'q' or event.key == 'escape':
            self.post_message(self.Escape())

    @property
    def latest_articles_list(self) -> SelectionList:
        return self.query_exactly_one(SelectionList)

    @on(Button.Pressed, '#summary_button')
    async def on_create_summary(self):
        self.create_summary()

    @on(Button.Pressed, '#open_button')
    async def on_open_article(self):
        selected_item_index = self.latest_articles_list.highlighted
        if selected_item_index:
            feed_entry: FeedEntry = self.current_items[cast(int, selected_item_index)]
            article = feed_entry.meta.base_article if feed_entry.meta.base_article \
                else self.feed_manager.extract_article(feed_entry)
            await self.app.push_screen(ArticleScreen(feed_entry, article))

    @work(exclusive=True, exit_on_error=False)
    async def create_summary(self) -> Article:
        selected_item_indexes = self.latest_articles_list.selected
        selected_entries = [self.current_items[index] for index in selected_item_indexes]
        return await articles_summary(selected_entries)

    def on_worker_state_changed(self, event: Worker.StateChanged) -> None:
        # only the worker completed and failed events are interesting
        if event.state == WorkerState.ERROR:
            self.notify(
                title="Failed to generate latest articles summary",
                message=f"Error: {str(event.worker.error)[:300]}",
                severity="error")
            self.log(event)
        elif event.state == WorkerState.SUCCESS:
            result = event.worker.result
            self.app.push_screen(ArticleScreen(article=result))

    @staticmethod
    def _selection_list_items(titles: list[str]) -> list[tuple[str, int, bool]]:
        return [(title, index, False) for index, title in enumerate(titles)]

    def _last_opened_time(self) -> datetime:
        # it's important to NOT return the min date here because then
        # generating a date that's prior to the min one, throws an exception
        # Default to ~ year ago
        return self.session_manager.session.last_opened \
            if self.session_manager.session.last_feed_refresh \
            else datetime.now() - timedelta(weeks=54)



from pydantic import BaseModel

from model.article import Article


class ArticleVersion(BaseModel):
    """
    An extension to the Article class that adds metadata regarding the
    reason for, and the options with which an article was generated.
    For instance, it's a translation of a source article, or it was a summary
    """
    parent_article: Article | None = None
    article: Article | None = None


class TranslatedArticle(ArticleVersion):
    language: str


class SummarizedArticle(ArticleVersion):
    target_word_length: int
